from keras.layers import Conv2D, Activation, Dense, Flatten, BatchNormalization, MaxPooling2D, AveragePooling2D, Dropout
from keras.engine import InputLayer
from keras.optimizers import Adadelta

from boarml.builders.base_builder import BaseBuilder
from keras import Sequential

from boarml.cnn import Layers
from boarml.cnn import Model


class KerasBuilder(BaseBuilder):

    def build(self, layers):
        model = Sequential()

        for layer in layers:
            model.add(self.create_layer(layer))

        opt = Adadelta()
        model.compile(optimizer=opt, loss='categorical_crossentropy', metrics=['accuracy'])

        return Model(model, 'keras')

    def create_layer(self, layer):
        layer_type = layer.type

        if layer_type == Layers.CONV:
            return Conv2D(filters=layer.filters, kernel_size=layer.kernel_size, strides=layer.strides, padding='same')
        elif layer_type == Layers.ACTIVATION:
            return Activation(layer.func)
        elif layer_type == Layers.INPUT:
            return InputLayer(input_shape=layer.shape)
        elif layer_type == Layers.DENSE:
            return Dense(units=layer.nodes)
        elif layer_type == Layers.FLATTEN:
            return Flatten()
        elif layer_type == Layers.DROPOUT:
            return Dropout(rate=layer.rate)
        elif layer_type == Layers.POOLING:
            if layer.operation == 'max':
                return MaxPooling2D(pool_size=layer.pool_size, strides=layer.strides, padding='same')
            elif layer.operation == 'average':
                return AveragePooling2D(pool_size=layer.pool_size, strides=layer.strides, padding='same')
        elif layer_type == Layers.BATCH_NORM:
            return BatchNormalization(momentum=layer.momentum)



class Unit:
    UNIT_SYMBOL = '-Unit-'

    def __init__(self):
        self.layers = list()

    def add_layer(self, layer):
        self.layers.append(layer)

    @property
    def size(self):
        return len(self.layers)

    def __str__(self):
        layer_count = len(self.layers)
        text = f'{Unit.UNIT_SYMBOL}({layer_count})'
        newline = '\n'

        for layer in self.layers:
            text = text + newline + str(layer)

        return text



class Model:

    def __init__(self, model, type):
        self.model = model
        self.type = type

    def fit(self, x_train, y_train, epochs, batch_size):
        if self.type == 'keras':
            return self.model.fit(x_train, y_train, epochs=epochs, batch_size=batch_size)

    def predict(self, x_test):
        if self.type == 'keras':
            return self.model.predict(x_test)
